//
// Copyright 2002 Sony Corporation 
//
// Permission to use, copy, modify, and redistribute this software for
// non-commercial use is hereby granted.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//

#include <string.h>
#include <OPENR/OSyslog.h>
#include <OPENR/OPENRAPI.h>
#include <OPENR/core_macro.h>
#include <ant.h>
#include <EndpointTypes.h>
#include <TCPEndpointMsg.h>
#include "ControlInfo.h"
#include "Network.h"

Network::Network () : networkState(NETWORK_IDLE),
		  connectionIndex(-1),
		  verboseMode(VERBOSE_DEFAULT)
{
}

OStatus
Network::DoInit(const OSystemEvent& event)
{
    OSYSDEBUG(("Network::DoInit()\n"));

    NEW_ALL_SUBJECT_AND_OBSERVER;
    REGISTER_ALL_ENTRY;
    SET_ALL_READY_AND_NOTIFY_ENTRY;

    return oSUCCESS;
}

OStatus
Network::DoStart(const OSystemEvent& event)
{
    OSYSDEBUG(("Network::DoStart()\n"));

    ipstackRef = antStackRef("IPStack");

    for (int index = 0; index < NETWORK_CONNECTION_MAX; index++) {
        OStatus result = InitTCPConnection(index);
        if (result != oSUCCESS) return oFAIL;
    }

    networkState = NETWORK_LISTEN;

    ENABLE_ALL_SUBJECT;
    ASSERT_READY_TO_ALL_OBSERVER;

    Listen(0);

    return oSUCCESS;
}

OStatus
Network::DoStop(const OSystemEvent& event)
{
    OSYSDEBUG(("Network::DoStop()\n"));

    networkState = NETWORK_IDLE;

    DISABLE_ALL_SUBJECT;
    DEASSERT_READY_TO_ALL_OBSERVER;

    return oSUCCESS;
}

OStatus
Network::DoDestroy(const OSystemEvent& event)
{
    DELETE_ALL_SUBJECT_AND_OBSERVER;

    return oSUCCESS;
}

int
from2digits( const char *buf )
{
    char number[ 3 ];
    memcpy( number, buf, 2 );
    number[ 2 ] = 0;
    return atoi( number );
}

int
from4digits( const char *buf )
{
    char number[ 5 ];
    memcpy( number, buf, 4 );
    number[ 4 ] = 0;
    return atoi( number );
}

void
Network::ProcessReceivedMessage( int index, const byte* buf, int size )
{
    ControlInfo ci;

    memset( &ci, 0, sizeof ci );
    if ( !memcmp( buf, "MO", 2 ) ) {
	ci.id = CtID_Monet;
	ci.iValue = from4digits( (char*) &buf[2] );
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "MS", 2 ) ) {
	ci.id = CtID_MonetSync;
	ci.iValue = from4digits( (char*) &buf[2] );
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "ST", 2 ) ) {
	ci.id = CtID_StartSTN;
	ci.iValue = from2digits( (char*) &buf[2] );
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "SB", 2 ) ) {
	ci.id = CtID_StopSTN;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "SR", 2 ) ) {
	ci.id = CtID_ReadSTN;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "SS", 2 ) ) {
	ci.id = CtID_GetCurrentStateSTN;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	connectionIndex = index;
    } else if ( !memcmp( buf, "GC", 2 ) ) {
	imageType = IMAGE_LAYER_C;
	connectionIndex = index;
        observer[obsImage]->AssertReady();
    } else if ( !memcmp( buf, "GL", 2 ) ) {
	imageType = IMAGE_LAYER_L;
	connectionIndex = index;
        observer[obsImage]->AssertReady();
    } else if ( !memcmp( buf, "GM", 2 ) ) {
	imageType = IMAGE_LAYER_M;
	connectionIndex = index;
        observer[obsImage]->AssertReady();
    } else if ( !memcmp( buf, "GI", 2 ) ) {
	ci.id = CtID_GetCdtInfo;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	connectionIndex = index;
    } else if ( !memcmp( buf, "CG", 2 ) ) {
	ci.id = CtID_GetCdt;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	connectionIndex = index;
    } else if ( !memcmp( buf, "CR", 2 ) ) {
	ci.id = CtID_ReadCdt;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "HD", 2 ) ) {
	// 0123456789012345678901234
	// HDXXX,XXX,-XX,-XX,-XX,-XX
	ci.id = CtID_Head;
	char arg[ 4 ];
	memcpy( arg, &buf[2], 3 );  arg[3] = '\0'; // XXX(Command)
	ci.iValue = atoi(arg);
	memcpy( arg, &buf[6], 3 );  arg[3] = '\0'; // XXX(Arg1)
	ci.i2Value = atoi(arg);
	memcpy( arg, &buf[10], 3 );  arg[3] = '\0'; // -XX(TILT)
	ci.dValue2[0] = atof(arg);
	memcpy( arg, &buf[14], 3 ); arg[3] = '\0'; // -XX(PAN)
	ci.dValue2[1] = atof(arg);
	memcpy( arg, &buf[18], 3 ); arg[3] = '\0'; // -XX(TILT2)
	ci.dValue2[2] = atof(arg);
	memcpy( arg, &buf[22], 3 ); arg[3] = '\0'; // -XX(MOUTH)
	ci.dValue2[3] = atof(arg);
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "HR", 2 ) ) {
	ci.id = CtID_ReadHead;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "FC", 2 ) ) {
	ci.id = CtID_Face;
	ci.iValue = from2digits( (char*) &buf[2] );
	if ( verboseMode >= 7 ) {
	    OSYSPRINT(( "%s FaceCmd=%d.\n",
			"Network::ProcessReceivedMessage", ci.iValue ));
	}
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "SP", 2 ) ) {
	// SP-XXX.XX,-XXX.XX,-XXX.XX,-XXX.XX,-XXX.XX,-XXX.XX,\\
	// -XXX.XX,-XXX.XX,-XXX.XX,-XXX.XX,-XXX.XX,-XXX.XX,
	ci.id = CtID_SetPose;
	const byte *ptr = &buf[2];
	char arg[10];
	for ( int i = 0; i < 12; i += 1 ) {
	    memcpy( arg, ptr, 7 );  // -XXX.XX
	    arg[7] = 0;
	    ci.dValue2[i] = atof(arg);
	    ptr += 8;
	}
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "SC", 2 ) ) {
	ci.id = CtID_ClearSTN;
	ci.iValue = from2digits( (char*) &buf[2] );
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "SA", 2 ) ) {
	ci.id = CtID_AddSTN;
	ci.iValue = from2digits( (char*) &buf[2] );
	strncpy( ci.sValue, (char*) &buf[4], sizeof ci.sValue );
	char *p = strchr( ci.sValue, '$' );
	if ( p ) {
	    *p = 0;
	}
	if ( verboseMode >= 7 ) {
	    OSYSPRINT(( "%s AddSTN=%-10.10s.\n",
			"Network::ProcessReceivedMessage", ci.sValue ));
	}
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "CS", 2 ) ) {
	ci.id = CtID_SendCdt;
	strncpy( ci.sValue, (char*) &buf[2], sizeof ci.sValue );
	char *p = strchr( ci.sValue, '$' );
	if ( p ) {
	    *p = 0;
	}
	if ( verboseMode >= 7 ) {
	    OSYSPRINT(( "%s SendCdt=%-10.10s.\n",
			"Network::ProcessReceivedMessage", ci.sValue ));
	}
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else if ( !memcmp( buf, "CW", 2 ) ) {
	ci.id = CtID_WriteCdt;
	subject[sbjControl]->SetData( &ci, sizeof(ci) );
	subject[sbjControl]->NotifyObservers();
	SendOkWithTagString( index, "OK", "" );
    } else {
	OSYSPRINT(( "%s unknown command: (%-10.10s)\n",
		    "Network::ProcessReceivedMessage",
		    buf ));
	SendOkWithTagString( index, "NG", "" );
    }
}

void
Network::NotifyImage(const ONotifyEvent& event)
{
    // Visionǡä
    if ( verboseMode >= 9 ) {
	OSYSPRINT(("Network::NotifyImage\n" ));
    }
    if ( networkState == NETWORK_IDLE ) {
        return;
    }
    if ( connectionIndex < 0 ) {
	return;
    }
    byte* data;

    switch ( imageType ) {
    case IMAGE_LAYER_C:
	data = (byte*) event.Data(0);
	SendOkWithTagData( connectionIndex, "IC", data, LAYER_C_IMAGESIZE );
	break;

    case IMAGE_LAYER_L:
	data = (byte*) event.Data(1);
	SendOkWithTagData( connectionIndex, "IL", data, LAYER_L_IMAGESIZE );
	break;

    case IMAGE_LAYER_M:
	data = (byte*) event.Data(2);
	SendOkWithTagData( connectionIndex, "IM", data, LAYER_M_IMAGESIZE );
	break;
    }
    connectionIndex = -1;
}

void
Network::NotifyControlResult( const ONotifyEvent& event )
{
    // Brainä
    if ( verboseMode >= 9 ) {
	OSYSPRINT(("Network::NotifyControlResult\n" ));
    }
    if ( networkState == NETWORK_IDLE ) {
        return;
    }
    if ( connectionIndex < 0 ) {
	return;
    }
    ControlInfo *cr = (ControlInfo*) event.Data(0);
    switch ( cr->id ) {
    case CtID_GetCurrentStateSTN:
	SendOkWithTagString( connectionIndex, "SS", cr->sValue );
	break;
    case CtID_GetCdtInfo:
	SendOkWithTagString( connectionIndex, "BI", cr->sValue );
	break;
    case CtID_GetCdt:
	SendOkWithTagString( connectionIndex, "CI", cr->sValue );
	break;
    }
    observer[event.ObsIndex()]->AssertReady();
    connectionIndex = -1;
}
