//
// Copyright 2002,2003 Sony Corporation 
//
// Permission to use, copy, modify, and redistribute this software for
// non-commercial use is hereby granted.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//

#include <string.h>
#include <OPENR/OSyslog.h>
#include <OPENR/OPENRAPI.h>
#include <OPENR/core_macro.h>
#include <ant.h>
#include <EndpointTypes.h>
#include <UDPEndpointMsg.h>
#include "NetworkUdp.h"
#include "entry.h"

static int ObjPositionToString( char *ptr, const ObjPosition *objPos );
static int FoundInfoToString( char *ptr, const char *tag, const FoundInfo *foundInfo );
static int STNInfoToString( char *ptr, const STNInfo *stnInfo );

NetworkUdp::NetworkUdp ()
{
    memset( &cdtInfo, 0, sizeof cdtInfo );
    memset( &stnInfo, 0, sizeof stnInfo );
    stnInfo.stnNo = -1;
}

OStatus
NetworkUdp::DoInit(const OSystemEvent& event)
{
    OSYSDEBUG(("NetworkUdp::DoInit()\n"));

    NEW_ALL_SUBJECT_AND_OBSERVER;
    REGISTER_ALL_ENTRY;
    SET_ALL_READY_AND_NOTIFY_ENTRY;

    ipstackRef = antStackRef("IPStack");

    for (int index = 0; index < NETWORKUDP_CONNECTION_MAX; index++) {
        OStatus result = InitUDPBuffer(index);
        if (result != oSUCCESS) return oFAIL;
    }

    return oSUCCESS;
}

OStatus
NetworkUdp::DoStart(const OSystemEvent& event)
{
    OStatus result;
    OSYSDEBUG(("NetworkUdp::DoStart()\n"));

    for (int index = 0; index < NETWORKUDP_CONNECTION_MAX; index++) {
        // 
        // Bind and Start receive data
        //
        result = CreateUDPEndpoint(index);
        if (result != oSUCCESS) {
            OSYSLOG1((osyslogERROR, "%s : %s[%d]",
                  "NetworkUdp::DoStart()",
                  "CreateUDPEndpoint() fail", index));
            continue;
        }
        result = Bind(index);
        if (result != oSUCCESS) {
            OSYSLOG1((osyslogERROR, "%s : %s[%d]",
                  "NetworkUdp::DoStart()",
                  "Bind() fail", index));
            continue;
        }
    }

    ENABLE_ALL_SUBJECT;
    ASSERT_READY_TO_ALL_OBSERVER;

    return oSUCCESS;
}    

OStatus
NetworkUdp::DoStop(const OSystemEvent& event)
{
    OSYSDEBUG(("NetworkUdp::DoStop()\n"));

    for (int index = 0; index < NETWORKUDP_CONNECTION_MAX; index++) {
        if (connection[index].state != CONNECTION_CLOSED &&
            connection[index].state != CONNECTION_CLOSING) {

            // Connection close
            UDPEndpointCloseMsg closeMsg(connection[index].endpoint);
            closeMsg.Call(ipstackRef, sizeof(closeMsg));
            connection[index].state = CONNECTION_CLOSED;
        }
    }

    DISABLE_ALL_SUBJECT;
    DEASSERT_READY_TO_ALL_OBSERVER;

    return oSUCCESS;
}

OStatus
NetworkUdp::DoDestroy(const OSystemEvent& event)
{
    for (int index = 0; index < NETWORKUDP_CONNECTION_MAX; index++) {
        // UnMap and Destroy SendBuffer
        connection[index].sendBuffer.UnMap();
        antEnvDestroySharedBufferMsg destroySendBufferMsg(connection[index].sendBuffer);
        destroySendBufferMsg.Call(ipstackRef, sizeof(destroySendBufferMsg));

        // UnMap and Destroy RecvBuffer
        connection[index].recvBuffer.UnMap();
        antEnvDestroySharedBufferMsg destroyRecvBufferMsg(connection[index].recvBuffer);
        destroyRecvBufferMsg.Call(ipstackRef, sizeof(destroyRecvBufferMsg));
    }

    DELETE_ALL_SUBJECT_AND_OBSERVER;

    return oSUCCESS;
}

OStatus
NetworkUdp::Send(int index)
{
    OSYSDEBUG(("NetworkUdp::Send()\n"));

    if (connection[index].sendSize == 0 ||
        connection[index].state != CONNECTION_CONNECTED) return oFAIL;

    UDPEndpointSendMsg sendMsg(connection[index].endpoint,
                               connection[index].sendAddress,
                               connection[index].sendPort,
                               connection[index].sendData,
                               connection[index].sendSize);
    sendMsg.continuation = (void*)index;
    sendMsg.Send(ipstackRef, myOID_,
                 Extra_Entry[entrySendCont], sizeof(UDPEndpointSendMsg));
                 

    connection[index].state = CONNECTION_SENDING;
    return oSUCCESS;
}

void
NetworkUdp::SendCont(ANTENVMSG msg)
{
    OSYSDEBUG(("NetworkUdp::SendCont()\n"));

    UDPEndpointSendMsg* sendMsg = (UDPEndpointSendMsg*)antEnvMsg::Receive(msg);
    int index = (int)(sendMsg->continuation);
    if (connection[index].state == CONNECTION_CLOSED)
        return;

    if (sendMsg->error != UDP_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s %d",
                  "NetworkUdp::SendCont()",
                  "FAILED. sendMsg->error", sendMsg->error));
        Close(index);
        return;
    }
    char in_buff[64];
    OSYSDEBUG(("sendData : %s", connection[index].sendData));
    OSYSDEBUG(("sendAddress : %s\n", connection[index].sendAddress.GetAsString(in_buff)));
    OSYSDEBUG(("sendPort : %d\n\n", connection[index].sendPort));
    
    connection[index].state = CONNECTION_CONNECTED;
    connection[index].recvSize = NETWORKUDP_BUFFER_SIZE;
    Receive(index);
}

OStatus
NetworkUdp::Receive(int index)
{
    OSYSDEBUG(("NetworkUdp::Receive()\n"));

    if (connection[index].state != CONNECTION_CONNECTED &&
        connection[index].state != CONNECTION_SENDING) return oFAIL;

    UDPEndpointReceiveMsg receiveMsg(connection[index].endpoint,
                                     connection[index].recvData,
                                     connection[index].recvSize);
    receiveMsg.continuation = (void*)index;

    receiveMsg.Send(ipstackRef, myOID_,
                    Extra_Entry[entryReceiveCont], sizeof(receiveMsg));

    return oSUCCESS;
}

void
NetworkUdp::ReceiveCont(ANTENVMSG msg)
{
    OSYSDEBUG(("NetworkUdp::ReceiveCont()\n"));

    UDPEndpointReceiveMsg* receiveMsg
        = (UDPEndpointReceiveMsg*)antEnvMsg::Receive(msg);

    int index = (int)(receiveMsg->continuation);
    if (connection[index].state == CONNECTION_CLOSED) return;

    if (receiveMsg->error != UDP_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s %d",
                  "NetworkUdp::ReceiveCont()",
                  "FAILED. receiveMsg->error", receiveMsg->error));
        Close(index);
        return;
    }

    OSYSDEBUG(("recvData : %s", receiveMsg->buffer));

    //
    // Send back the message
    //
    connection[index].sendAddress = receiveMsg->address;
    connection[index].sendPort    = receiveMsg->port;

    char *ptr = (char*) connection[index].sendData ;
    *ptr++ = 'S';
    *ptr++ = '#';
    ptr += STNInfoToString( ptr, &stnInfo );
    ptr += FoundInfoToString( ptr, "BL", &cdtInfo.foundInfo[IO_BALL] );
    ptr += FoundInfoToString( ptr, "GB", &cdtInfo.foundInfo[IO_GOAL_B] );
    ptr += FoundInfoToString( ptr, "GY", &cdtInfo.foundInfo[IO_GOAL_Y] );
    ptr += FoundInfoToString( ptr, "PB", &cdtInfo.foundInfo[IO_BEACON_PB] );
    ptr += FoundInfoToString( ptr, "BP", &cdtInfo.foundInfo[IO_BEACON_BP] );
    ptr += FoundInfoToString( ptr, "PY", &cdtInfo.foundInfo[IO_BEACON_PY] );
    ptr += FoundInfoToString( ptr, "YP", &cdtInfo.foundInfo[IO_BEACON_YP] );
    *ptr = '\0';
    connection[index].sendSize    = strlen((char*)connection[index].sendData);

    Send(index);
}

OStatus
NetworkUdp::Close(int index)
{
    OSYSDEBUG(("NetworkUdp::Close()\n"));

    if (connection[index].state == CONNECTION_CLOSED ||
        connection[index].state == CONNECTION_CLOSING) return oFAIL;

    UDPEndpointCloseMsg closeMsg(connection[index].endpoint);
    closeMsg.continuation = (void*)index;

    closeMsg.Send(ipstackRef, myOID_,
                  Extra_Entry[entryCloseCont], sizeof(closeMsg));

    connection[index].state = CONNECTION_CLOSING;

    return oSUCCESS;
}

void
NetworkUdp::CloseCont(ANTENVMSG msg)
{
    OStatus result;
    OSYSDEBUG(("NetworkUdp::CloseCont()\n"));
    
    UDPEndpointCloseMsg* closeMsg = (UDPEndpointCloseMsg*)antEnvMsg::Receive(msg);
    int index = (int)(closeMsg->continuation);
    if (connection[index].state == CONNECTION_CLOSED)
        return;

    connection[index].state = CONNECTION_CLOSED;

    result = CreateUDPEndpoint(index);
    if (result != oSUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d]",
                  "NetworkUdp::CloseCont()",
                  "CreateUDPEndpoint() fail", index));
        return;
    }
    result = Bind(index);
    if (result != oSUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d]",
                  "NetworkUdp::CloseCont()",
                  "Bind() fail", index));
    }
}

OStatus
NetworkUdp::InitUDPBuffer(int index)
{
    OSYSDEBUG(("NetworkUdp::InitUDPBuffer()\n"));

    connection[index].state = CONNECTION_CLOSED;

    // 
    // Allocate send buffer
    //
    antEnvCreateSharedBufferMsg sendBufferMsg(NETWORKUDP_BUFFER_SIZE);

    sendBufferMsg.Call(ipstackRef, sizeof(sendBufferMsg));
    if (sendBufferMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "NetworkUdp::InitUDPBuffer()",
                  "Can't allocate send buffer",
                  index, sendBufferMsg.error));
        return oFAIL;
    }

    connection[index].sendBuffer = sendBufferMsg.buffer;
    connection[index].sendBuffer.Map();
    connection[index].sendData
        = (byte*)(connection[index].sendBuffer.GetAddress());

    //
    // Allocate receive buffer
    //
    antEnvCreateSharedBufferMsg recvBufferMsg(NETWORKUDP_BUFFER_SIZE);

    recvBufferMsg.Call(ipstackRef, sizeof(recvBufferMsg));
    if (recvBufferMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "NetworkUdp::InitUDPBuffer()",
                  "Can't allocate receive buffer",
                  index, recvBufferMsg.error));
        return oFAIL;
    }

    connection[index].recvBuffer = recvBufferMsg.buffer;
    connection[index].recvBuffer.Map();
    connection[index].recvData
        = (byte*)(connection[index].recvBuffer.GetAddress());
    connection[index].recvSize = NETWORKUDP_BUFFER_SIZE;

    return oSUCCESS;
}

OStatus
NetworkUdp::CreateUDPEndpoint(int index)
{
    OSYSDEBUG(("NetworkUdp::CreateUDPEndpoint()\n"));

    if (connection[index].state != CONNECTION_CLOSED) return oFAIL;

    //
    // Create UDP endpoint
    //
    antEnvCreateEndpointMsg udpCreateMsg(EndpointType_UDP,
                                         NETWORKUDP_BUFFER_SIZE * 2);
    udpCreateMsg.Call(ipstackRef, sizeof(udpCreateMsg));
    if (udpCreateMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "NetworkUdp::CreateUDPEndpoint()",
                  "Can't create endpoint",
                  index, udpCreateMsg.error));
        return oFAIL;
    }
    connection[index].endpoint = udpCreateMsg.moduleRef;

    return oSUCCESS;
}


OStatus
NetworkUdp::Bind(int index)
{
    OSYSDEBUG(("NetworkUdp::Bind()\n"));

    if (connection[index].state != CONNECTION_CLOSED) return oFAIL;

    // 
    // Bind
    //
    UDPEndpointBindMsg bindMsg(connection[index].endpoint, 
                               IP_ADDR_ANY, NETWORKUDP_PORT);
    bindMsg.Call(ipstackRef,sizeof(antEnvCreateEndpointMsg));
    if (bindMsg.error != UDP_SUCCESS) {
        return oFAIL;
    }

    connection[index].state = CONNECTION_CONNECTED;
    connection[index].recvSize = NETWORKUDP_BUFFER_SIZE;

    Receive(index);

    return oSUCCESS;
}

void
NetworkUdp::NotifyObjInfo( const ONotifyEvent& event )
{
    // Visionܡä
    //OSYSPRINT(("NetworkUdp::NotifyObjInfo\n"));
    cdtInfo = *(CdtInfo*) event.Data(0);
    observer[event.ObsIndex()]->AssertReady();
}

void
NetworkUdp::NotifySTNInfo( const ONotifyEvent& event )
{
    // BrainSTNä
    stnInfo = *(STNInfo*) event.Data(0);
    //OSYSPRINT(("NetworkUdp::NotifySTNInfo(%d)\n", stnInfo.stnNo));
    observer[event.ObsIndex()]->AssertReady();
}

static int
ObjPositionToString( char *ptr, const ObjPosition *objPos )
{
    sprintf( ptr, "%03d%03d%03d%03d%06.1f%06.1f",
	     objPos->x, objPos->y, objPos->w, objPos->h,
	     objPos->pan, objPos->tilt );
    return 3*4 + 6*2;
}

static int
FoundInfoToString( char *ptr, const char *tag, const FoundInfo *foundInfo )
{
    int size = 0;
    sprintf( &ptr[size], "%s", tag );
    size += 2;
    sprintf( &ptr[size], "%01d", foundInfo->found ? 1 : 0 );
    size += 1;
    size += ObjPositionToString( &ptr[size], &foundInfo->objPos );
    sprintf( &ptr[size], "%03d:", foundInfo->distance % 1000 );
    size += 4;
    return size;
}

static int
STNInfoToString( char *ptr, const STNInfo *stnInfo )
{
    sprintf( ptr, "!%02d", stnInfo->stnNo ); ptr += 3;
    for ( int i = 0; i < MAX_TASKS; i ++ ) {
	sprintf( ptr, "%02d", stnInfo->stateIds[i] ); ptr += 2;
    }
    *ptr++ = ':';
    *ptr = 0;
    return 4 + MAX_TASKS * 2;
}
